// Multi-set mapping expression -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "Mapping.h"
#include "VariableDefinition.h"
#include "Valuation.h"
#include "PlaceMarking.h"
#include "LeafValue.h"
#include "Substitution.h"
#include "Variable.h"
#include "Printer.h"

/** @file Mapping.C
 * Multi-set mapping operation
 */

Mapping::Mapping (class VariableDefinition& variable,
		  class VariableDefinition* cardinality,
		  class Expression& marking,
		  class Expression& expr) :
  myVariable (&variable), myCardinality (cardinality),
  myMarking (&marking), myExpr (&expr)
{
  assert (myVariable && myExpr);
  assert (myMarking && myMarking->getType ());
  assert (!myMarking->isTemporal () && myExpr->isBasic ());
  if (myCardinality) {
    assert (myCardinality->getType ().getKind () == Type::tCard);
    setType (*myMarking->getType ());
  }
  else
    setType (*myExpr->getType ());
}

Mapping::~Mapping ()
{
  delete myVariable;
  delete myCardinality;
  myMarking->destroy ();
  myExpr->destroy ();
}

class PlaceMarking*
Mapping::meval (const class Valuation& valuation) const
{
  if (class PlaceMarking* p = myMarking->meval (valuation)) {
    p->setPlace (NULL);
    class Valuation v (valuation);
    if (myCardinality) {
      for (PlaceMarking::iterator i = p->begin (); i != p->end (); i++) {
	v.setValue (*myVariable, *PlaceMarking::getValue (i).copy ());
	v.setValue (*myCardinality,
		    *new class LeafValue (myCardinality->getType (),
					  PlaceMarking::getCount (i)));
	if (class Value* val = myExpr->eval (v)) {
	  assert (val->getKind () == Value::vLeaf);
	  card_t cnt = card_t (static_cast<class LeafValue&>(*val));
	  delete val;
	  p->setCount (i, cnt);
	}
	else {
	  valuation.copyErrors (v);
	  delete p;
	  return NULL;
	}
      }
      return p;
    }
    else {
      class PlaceMarking* q = new class PlaceMarking ();
#ifndef NDEBUG
      q->setType (myExpr->getType ());
#endif // NDEBUG

      for (PlaceMarking::iterator i = p->begin (); i != p->end (); i++) {
	if (!PlaceMarking::getCount (i))
	  continue;
	v.setValue (*myVariable, *PlaceMarking::getValue (i).copy ());
	if (class Value* val = myExpr->eval (v)) {
	  if (!q->add (*val, PlaceMarking::getCount (i))) {
	    valuation.flag (errCard, *this);
	    delete p;
	    delete q;
	    return NULL;
	  }
	}
	else {
	  valuation.copyErrors (v);
	  delete p;
	  delete q;
	  return NULL;
	}
      }

      delete p;
      return q;
    }
  }
  return NULL;
}

class Expression*
Mapping::ground (const class Valuation& valuation,
		 class Transition* transition,
		 bool declare)
{
  class Expression* marking =
    myMarking->ground (valuation, transition, declare);
  if (!marking) return NULL;
  class Expression* expr =
    myExpr->ground (valuation, transition, declare);
  if (!expr) { marking->destroy (); return NULL; }

  assert (valuation.isOK ());

  if (marking == myMarking && expr == myExpr) {
    marking->destroy ();
    expr->destroy ();
    return copy ();
  }
  else {
    class VariableDefinition* variable =
      new class VariableDefinition (*myVariable);
    class VariableDefinition* cardinality = myCardinality
      ? new class VariableDefinition (*myCardinality)
      : NULL;
    class Substitution subs;
    subs.setExpr (*myVariable, *new class Variable (*variable));
    if (myCardinality)
      subs.setExpr (*myCardinality, *new class Variable (*cardinality));
    return new class Mapping (*variable, cardinality,
			      *marking->substitute (subs),
			      *expr->substitute (subs));
  }
}

class Expression*
Mapping::substitute (class Substitution& substitution)
{
  class Expression* marking = myMarking->substitute (substitution);
  class Expression* expr = myExpr->substitute (substitution);

  if (!marking || !expr) {
    marking->destroy ();
    expr->destroy ();
    return NULL;
  }
  if (marking == myMarking && expr == myExpr) {
    marking->destroy ();
    expr->destroy ();
    return copy ();
  }
  else
    return new class Mapping (*myVariable, myCardinality, *marking, *expr);
}

bool
Mapping::depends (const class VariableSet& vars,
		  bool complement) const
{
  return
    myMarking->depends (vars, complement) ||
    myExpr->depends (vars, complement);
}

bool
Mapping::forVariables (bool (*operation)
		       (const class Expression&,void*),
		       void* data) const
{
  return
    myMarking->forVariables (operation, data) &&
    myExpr->forVariables (operation, data);
}

#ifdef EXPR_COMPILE
# include "CExpression.h"
# include "Constant.h"
# include "PlaceContents.h"
# include "Place.h"

void
Mapping::compileMset (class CExpression& cexpr,
		      unsigned indent,
		      const char* result,
		      const class VariableSet* vars) const
{
  assert (myVariable && myMarking && myExpr);
  /** the multi-set iterator */
  char* iter;
  /** the mapping expression */
  char* expr;
  /** the item iterator */
  char* var;
  /** the multiplicity iterator */
  char* card;
  if (myExpr->getKind () == Expression::eConstant) {
    if (cexpr.getVariable (static_cast<const class Constant&>(*myExpr), expr))
      myExpr->compile (cexpr, indent, expr, vars);
    var = card = 0;
  }
  else {
    cexpr.getVariable (*myExpr, expr);
    var = cexpr.getIterator (*myVariable);
    card = myCardinality ? cexpr.getIterator (*myCardinality) : 0;
  }

  class StringBuffer& out = cexpr.getOut ();
  /** flag: is the marking the contents of a place with at most one token? */
  bool place1 = myMarking->getKind () == Expression::ePlaceContents &&
    static_cast<const class PlaceContents*>
    (myMarking)->getPlace ().getMaxNumTokens () == 1;

  if (place1) {
    iter = static_cast<const class PlaceContents*>(myMarking)->getName (cexpr);
    out.indent (indent);
    out.append ("if ("), out.append (iter), out.append (") {\n");
    if (myExpr->getKind () != Expression::eConstant) {
      out.indent (indent + 2);
      myVariable->getType ().appendName (out);
      out.append (" "), out.append (var), out.append (" = *");
      out.append (iter);
      out.append (";\n");
      if (card) {
	out.indent (indent + 2);
	out.append ("card_t "), out.append (card), out.append ("=1;\n");
      }
      myExpr->compile (cexpr, indent + 2, expr, vars);
    }
    out.indent (indent + 2);
  }
  else {
    /** the multi-set */
    char* mset;
    if (cexpr.getVariable (*myMarking, mset))
      myMarking->compileMset (cexpr, indent, mset, vars);
    iter = cexpr.getLabel ();
    out.indent (indent), out.append ("{\n");
    out.indent (indent + 2);
    myMarking->getType ()->appendMSetName (out);
    out.append ("* "), out.append (iter), out.append (" = ");
    out.append (mset), out.append (";\n");
    delete[] mset;
    out.indent (indent + 2);
    out.append ("FIRST ("), out.append (iter), out.append (");\n");
    out.indent (indent + 2);
    out.append ("while ("), out.append (iter), out.append (") {\n");
    out.indent (indent + 4);
    out.append ("if ("), out.append (iter), out.append ("->count) {\n");
    if (myExpr->getKind () != Expression::eConstant) {
      out.indent (indent + 6);
      myVariable->getType ().appendName (out);
      out.append (" "), out.append (var), out.append ("=");
      out.append (iter), out.append ("->item;\n");
      if (card) {
	out.indent (indent + 6);
	out.append ("card_t "), out.append (card), out.append ("=");
	out.append (iter), out.append ("->count;\n");
      }
      myExpr->compile (cexpr, indent + 6, expr, vars);
    }
    out.indent (indent + 6);
  }
  out.append (result);
  out.append ("=insert");
  getType ()->appendIndex (out);
  out.append (" ("), out.append (result), out.append (", ");
  if (card) {
    if (place1)
      out.append (iter), out.append (", ");
    else
      out.append ("&"), out.append (iter), out.append ("->item, ");
    out.append (expr);
  }
  else
    out.append ("&"), out.append (expr), out.append (", 1");
  out.append (");\n");
  if (!place1) {
    out.indent (indent + 4), out.append ("}\n");
    out.indent (indent + 4);
    out.append ("NEXT ("), out.append (iter), out.append (");\n");
    out.indent (indent + 2), out.append ("}\n");
  }
  out.indent (indent), out.append ("}\n");

  delete[] card;
  delete[] var;
  delete[] expr;
  delete[] iter;
}

#endif // EXPR_COMPILE

void
Mapping::display (const class Printer& printer) const
{
  printer.printRaw ("map");
  printer.delimiter (' ');
  printer.print (myVariable->getName ());
  if (myCardinality) {
    printer.delimiter ('#');
    printer.print (myCardinality->getName ());
  }
  printer.delimiter ('{')++;
  myMarking->display (printer);
  --printer.delimiter ('}');
  printer.delimiter ('(')++;
  myExpr->display (printer);
  --printer.delimiter (')');
}
