// Maria structure assignment class -*- c++ -*-

#ifndef STRUCTASSIGN_H_
# define STRUCTASSIGN_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "Expression.h"

/** @file StructAssign.h
 * Structure component assignment operation
 */

/** Structure component assignment */
class StructAssign : public Expression
{
public:
  /** Constructor
   * @param structure	the structure
   * @param index	index of the component to be assigned to
   * @param expr	the value to assign to the struct component
   */
  StructAssign (class Expression& structure,
		card_t index,
		class Expression& expr);
private:
  /** Copy constructor */
  StructAssign (const class StructAssign& old);
  /** Assignment operator */
  class StructAssign& operator= (const class StructAssign& old);
protected:
  /** Destructor */
  ~StructAssign ();
public:

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eStructAssign; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Equality comparison operator */
  bool operator== (const class StructAssign& other) const {
    return myIndex == other.myIndex &&
      myStructure == other.myStructure &&
      myExpr == other.myExpr;
  }

  /** Ordering comparison operator */
  bool operator< (const class StructAssign& other) const {
    if (myIndex < other.myIndex) return true;
    if (other.myIndex < myIndex) return false;
    if (myStructure < other.myStructure) return true;
    if (other.myStructure < myStructure) return false;
    return myExpr < other.myExpr;
  }

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation& valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @param transition	Transition for registering quantified variables
   * @param declare	flag: declare new variables if required
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation,
			    class Transition* transition,
			    bool declare);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

  /** Determine whether the expression depends on a set of variables
   * @param vars	the set of variables
   * @param complement	flag: treat the set as its complement
   */
  bool depends (const class VariableSet& vars,
		bool complement) const;

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @param data	parameters to be passed to the operation
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class Expression&,void*),
		     void* data) const;

# ifdef EXPR_COMPILE
  /** Generate C code for evaluating the expression
   * @param cexpr	the compilation
   * @param indent	indentation level
   * @param lvalue	C expression referring to the lvalue
   * @param vars	the variables that have been assigned a value
   */
  void compile (class CExpression& cexpr,
		unsigned indent,
		const char* lvalue,
		const class VariableSet* vars) const;
# endif // EXPR_COMPILE

  /** Display the expression
   * @param printer	the printer object
   */
  void display (const class Printer& printer) const;

private:
  /** The structure */
  class Expression& myStructure;
  /** Index of the component to be assigned to */
  card_t myIndex;
  /** The replacement for the component */
  class Expression& myExpr;
};

#endif // STRUCTASSIGN_H_
