// Maria vector component assignment class -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "VectorAssign.h"
#include "VectorValue.h"
#include "VectorType.h"
#include "Constant.h"
#include "Printer.h"
#include <cstring>

/** @file VectorAssign.C
 * Vector component assignment operation
 */

VectorAssign::VectorAssign (class Expression& vect,
			    class Expression& index,
			    class Expression& item) :
  Expression (),
  myVector (vect), myIndex (index), myItem (item)
{
  assert (myVector.getType () &&
	  myVector.getType ()->getKind () == Type::tVector);
  assert (myVector.isBasic () && myIndex.isBasic () && myItem.isBasic ());
  const class VectorType& type =
    *static_cast<const class VectorType*>(myVector.getType ());
  assert (myItem.getType () && myItem.getType () == &type.getItemType ());
  assert (myIndex.getType () && myIndex.getType () == &type.getIndexType ());
  setType (type);
}

VectorAssign::~VectorAssign ()
{
  myVector.destroy ();
  myIndex.destroy ();
  myItem.destroy ();
}

class Value*
VectorAssign::do_eval (const class Valuation& valuation) const
{
  class Value* index = myIndex.eval (valuation);
  if (!index)
    return NULL;
  const class Type& indexType =
    static_cast<const class VectorType*>(myVector.getType ())->getIndexType ();

  assert (index->getType ().isAssignable (indexType));
  assert (indexType.isConstrained (*index));

  card_t i = indexType.convert (*index);
  assert (i < CARD_T_MAX && i < indexType.getNumValues ());
  delete index;

  class Value* vector = myVector.eval (valuation);
  if (!vector)
    return NULL;
  assert (&vector->getType () == myVector.getType ());
  class Value*& v = static_cast<class VectorValue&>(*vector)[i];
  delete v;
  if (!(v = myItem.eval (valuation))) {
    delete vector;
    return NULL;
  }
  return constrain (valuation, vector);
}

class Expression*
VectorAssign::ground (const class Valuation& valuation,
		      class Transition* transition,
		      bool declare)
{
  class Expression* vect = myVector.ground (valuation, transition, declare);
  if (!vect) return NULL;
  class Expression* index = myIndex.ground (valuation, transition, declare);
  if (!index) { vect->destroy (); return NULL; }
  class Expression* item = myItem.ground (valuation, transition, declare);
  if (!item) { vect->destroy (); index->destroy (); return NULL; }

  assert (valuation.isOK ());

  if (vect == &myVector && index == &myIndex && item == &myItem) {
    vect->destroy (), index->destroy (), item->destroy ();
    return copy ();
  }
  else {
    vect = new class VectorAssign (*vect, *index, *item);
    if (!vect->isBasic ());
    else if (class Value* v = vect->eval (valuation)) {
      vect->destroy ();
      return new class Constant (*v);
    }
    else if (!valuation.isOKorVar ()) {
      vect->destroy ();
      return NULL;
    }

    valuation.clearErrors ();
    return vect;
  }
}

class Expression*
VectorAssign::substitute (class Substitution& substitution)
{
  class Expression* vect = myVector.substitute (substitution);
  class Expression* index = myIndex.substitute (substitution);
  class Expression* item = myItem.substitute (substitution);

  if (vect == &myVector && index == &myIndex && item == &myItem) {
    vect->destroy (), index->destroy (), item->destroy ();
    return copy ();
  }
  else
    return new class VectorAssign (*vect, *index, *item);
}

bool
VectorAssign::depends (const class VariableSet& vars,
		       bool complement) const
{
  return
    myVector.depends (vars, complement) ||
    myIndex.depends (vars, complement) ||
    myItem.depends (vars, complement);
}

bool
VectorAssign::forVariables (bool (*operation)
			    (const class Expression&,void*),
			    void* data) const
{
  return
    myVector.forVariables (operation, data) &&
    myIndex.forVariables (operation, data) &&
    myItem.forVariables (operation, data);
}

#ifdef EXPR_COMPILE
# include "CExpression.h"

void
VectorAssign::compile (class CExpression& cexpr,
		       unsigned indent,
		       const char* lvalue,
		       const class VariableSet* vars) const
{
  myVector.compile (cexpr, indent, lvalue, vars);
  char* ixconv = 0;
  class StringBuffer& out = cexpr.getOut ();
  if (myIndex.getKind () != Expression::eConstant) {
    char* ixvalue = 0;
    if (cexpr.getVariable (myIndex, ixvalue))
      myIndex.compile (cexpr, indent, ixvalue, vars);
    if (cexpr.getConverted (myIndex, ixconv))
      myIndex.getType ()->compileConversion (out, indent,
					     ixvalue, ixconv, false);
    delete[] ixvalue;
  }
  class StringBuffer item;
  item.append (lvalue);
  item.append (".a[");
  if (myIndex.getKind () == Expression::eConstant)
    item.append (myIndex.getType ()->convert
		 (static_cast<const class Constant&>(myIndex).getValue ()));
  else
    item.append (ixconv);
  item.append ("]");
  delete[] ixconv;
  myItem.compile (cexpr, indent, item.getString (), vars);

  compileConstraint (cexpr, indent, lvalue);
}

#endif // EXPR_COMPILE

/** Determine whether an expression needs to be enclosed in parentheses
 * @param kind	kind of the expression
 * @return	whether parentheses are necessary
 */
inline static bool
needParentheses (enum Expression::Kind kind)
{
  switch (kind) {
  case Expression::eVariable:
  case Expression::eUndefined:
  case Expression::eStructComponent:
  case Expression::eUnionComponent:
  case Expression::eVectorIndex:
  case Expression::eVectorAssign:
    return false;
  case Expression::eCardinality:
  case Expression::eBuffer:
  case Expression::eBufferRemove:
  case Expression::eBufferWrite:
  case Expression::eBufferIndex:
  case Expression::eBinop:
  case Expression::eBooleanBinop:
  case Expression::eNot:
  case Expression::eRelop:
  case Expression::eSet:
  case Expression::eTemporalBinop:
  case Expression::eTemporalUnop:
  case Expression::eMarking:
  case Expression::eTransitionQualifier:
  case Expression::ePlaceContents:
  case Expression::eSubmarking:
  case Expression::eMapping:
  case Expression::eEmptySet:
  case Expression::eUnionType:
  case Expression::eUnion:
  case Expression::eStruct:
  case Expression::eStructAssign:
    assert (false);
  case Expression::eUnop:
  case Expression::eBufferUnop:
  case Expression::eConstant:
  case Expression::eIfThenElse:
  case Expression::eVector:
  case Expression::eVectorShift:
  case Expression::eTypecast:
    break;
  }

  return true;
}

void
VectorAssign::display (const class Printer& printer) const
{
  if (::needParentheses (myVector.getKind ())) {
    printer.delimiter ('(')++;
    myVector.display (printer);
    --printer.delimiter (')');
  }
  else
    myVector.display (printer);
  printer.delimiter ('.');
  printer.delimiter ('{')++;
  printer.delimiter ('[')++;
  myIndex.display (printer);
  --printer.delimiter (']');
  myItem.display (printer);
  --printer.delimiter ('}');
}
