// Maria vector indexing expression -*- c++ -*-

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__
#include "VectorIndex.h"
#include "VectorType.h"
#include "VectorValue.h"
#include "Valuation.h"
#include "Constant.h"
#include "VariableDefinition.h"
#include "Printer.h"

/** @file VectorIndex.C
 * Vector indexing operation
 */

VectorIndex::VectorIndex (class Expression& vector,
			  class Expression& index) :
  Expression (),
  myVector (&vector), myIndex (&index)
{
  assert (myVector && myVector->getType ());
  assert (myVector->getType ()->getKind () == Type::tVector);
  assert (myIndex && myIndex->getType ());
  assert (myIndex->getType ()->isAssignable
	  (static_cast<const class VectorType*>(myVector->getType ())
	   ->getIndexType ()));
  assert (myVector->isBasic () && myIndex->isBasic ());
  setType (static_cast<const class VectorType*>(myVector->getType ())
	   ->getItemType ());
}

VectorIndex::~VectorIndex ()
{
  myVector->destroy ();
  myIndex->destroy ();
}

class Value*
VectorIndex::do_eval (const class Valuation& valuation) const
{
  class Value* index = myIndex->eval (valuation);
  if (!index)
    return NULL;
  const class Type& indexType =
    static_cast<const class VectorType*>(myVector->getType ())
    ->getIndexType ();

  assert (index->getType ().isAssignable (indexType));
  assert (indexType.isConstrained (*index));

  card_t i = indexType.convert (*index);
  assert (i < CARD_T_MAX && i < indexType.getNumValues ());
  delete index;

  class Value* vector = myVector->eval (valuation);
  if (!vector)
    return NULL;

  assert (&vector->getType () == myVector->getType ());
  if (const class Value* v =
      (*static_cast<class VectorValue*>(vector))[i]) {
    class Value* w = v->copy ();
    delete vector;
    return w;
  }
  else {
    assert (false);
    delete vector;
    return NULL;
  }
}

class Expression*
VectorIndex::ground (const class Valuation& valuation,
		     class Transition* transition,
		     bool declare)
{
  class Expression* vector = myVector->ground (valuation, transition, declare);
  if (!vector) return NULL;
  class Expression* index = myIndex->ground (valuation, transition, declare);
  if (!index) { vector->destroy (); return NULL; }

  assert (valuation.isOK ());

  if (vector == myVector && index == myIndex) {
    vector->destroy ();
    index->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class VectorIndex (*vector, *index);
    if (!expr->isBasic ());
    else if (class Value* v = expr->eval (valuation)) {
      expr->destroy ();
      return new class Constant (*v);
    }
    else if (!valuation.isOKorVar ()) {
      expr->destroy ();
      return NULL;
    }

    valuation.clearErrors ();
    return expr;
  }
}

class Expression*
VectorIndex::substitute (class Substitution& substitution)
{
  class Expression* vector = myVector->substitute (substitution);
  class Expression* index = myIndex->substitute (substitution);

  if (vector == myVector && index == myIndex) {
    vector->destroy ();
    index->destroy ();
    return copy ();
  }
  else
    return new class VectorIndex (*vector, *index);
}

bool
VectorIndex::depends (const class VariableSet& vars,
		      bool complement) const
{
  return
    myVector->depends (vars, complement) ||
    myIndex->depends (vars, complement);
}

bool
VectorIndex::forVariables (bool (*operation)
			   (const class Expression&,void*),
			   void* data) const
{
  return
    myVector->forVariables (operation, data) &&
    myIndex->forVariables (operation, data);
}

#ifdef EXPR_COMPILE
# include "CExpression.h"

void
VectorIndex::compile (class CExpression& cexpr,
		      unsigned indent,
		      const char* lvalue,
		      const class VariableSet* vars) const
{
  char* vvalue = 0;
  if (cexpr.getVariable (*myVector, vvalue))
    myVector->compile (cexpr, indent, vvalue, vars);
  char* ixconv = 0;
  class StringBuffer& out = cexpr.getOut ();
  if (myIndex->getKind () != Expression::eConstant) {
    char* ixvalue = 0;
    if (cexpr.getVariable (*myIndex, ixvalue))
      myIndex->compile (cexpr, indent, ixvalue, vars);
    if (cexpr.getConverted (*myIndex, ixconv))
      myIndex->getType ()->compileConversion (out, indent,
					      ixvalue, ixconv, false);
    delete[] ixvalue;
  }
  out.indent (indent);
  out.append (lvalue);
  out.append ("=");
  out.append (vvalue);
  out.append (".a[");
  if (myIndex->getKind () == Expression::eConstant)
    out.append (myIndex->getType ()->convert
		(static_cast<const class Constant&>(*myIndex).getValue ()));
  else
    out.append (ixconv);
  out.append ("];\n");
  delete[] vvalue;
  delete[] ixconv;
}

#endif // EXPR_COMPILE

void
VectorIndex::display (const class Printer& printer) const
{
  myVector->display (printer);
  printer.delimiter ('[')++;
  myIndex->display (printer);
  --printer.delimiter (']');
}
